import numpy as np
import vegas_params as vp
#from vegas_params import expression, Expression, Uniform, Fixed, FromDistribution
from vegas_params.vector import vector
from vegas_params.utils import swapaxes, swap_axes
from vegas_params.utils import mask_arrays, unmask_arrays, save_input
#from vegas_params import integral

from concurrent.futures import ProcessPoolExecutor
from itertools import product
from functools import partial

from .sources import Point #dataclass which holds trajectory points
from .steps import StepsUniform, StepsDistribution
from .medium import Medium
from .utils import combine_rotations, align_rotation_to_vector

from loguru import logger

# --- Expressions are defined here
class Process(vp.Expression):
    r"""The calculator of the RTE term :math:`\delta L^{(n)}` of the orrder `Nsteps`:

    .. math ::
        \delta L^{(n)} = 
        A_{src}(\vec{R}_{src}, T_{src}, \hat{s}_{src}) \cdot A_{det}(\vec{R}_{det}, T_{det}, \hat{s}_{det}) \cdot e^{-\mu_t\,c\,(T_{det}-T_{src})}\cdot \left[\mu_s\,c\,(T_{det}-T_{src})\right]^n \cdot \prod_{i=1}^{n} f(\hat{s}_i\cdot\hat{s}_{i-1})\cdot \delta^3\left(\vec{R}_{tgt}-\vec{R}_{src}-\sum_{i=1}^{n} \hat{s}_{i} c (t_{i+1}-t_{i})\right)
    """
    def __init__(self, 
                 src:vp.Expression, 
                 tgt:vp.Expression, 
                 medium:Medium, 
                 Nsteps:int,
                 save_trajectory:bool=False, 
                 use_masking:bool=True,
                 use_uniform_sampling=False
                ):
        self.medium = medium
        self.Nsteps = Nsteps
        self.save_trajectory = save_trajectory
        self.use_masking = use_masking
        self.vegas_kwargs = {'nitn':10, 'neval':3000}

        if use_uniform_sampling:
            super().__init__(src=src,tgt=tgt,
                             steps=StepsUniform(Nsteps,medium)
                            )
        else:
            super().__init__(src=src,tgt=tgt,
                             steps=StepsDistribution(Nsteps,medium)
                            )

    def calculate_last_step(self, R, T, t_i, s_i):
        """calculate the length and direction of the last step, needed to get to the given point R,T"""
        dt_i = np.diff(t_i, axis=0)
        N = np.sum(s_i*dt_i, axis=0, keepdims=True).view(vector)
        R1 = R-N*T*self.medium.c
        R1_2 = R1.mag2()
        gamma = 1-N.mag2()
        R1N = R1.dot(N)
        #calculate the missing step length
        #the equation is 
        # gamma*(l**2) - 2*R1N*l - R1**2 = 0
        is_linear = np.isclose(gamma, 0)#in these cases the solution is linear
        #determinant
        D = R1N**2+R1_2*gamma
        
        step_length = np.where(is_linear, -R1_2/(2*R1N), (R1N+np.sqrt(D))/(gamma))
        step_direction = R1/step_length+N
        #calculate factor
        factor = 1/(step_length*np.abs(R1N - step_length*gamma))
        factor *= 1/T
        factor *= 1/self.medium.c
        return step_length, step_direction, factor
        
    @swapaxes
    def __call__(self, src, tgt, steps):
        #target points
        R = tgt.R-src.R
        T = tgt.T-src.T
        #convenience stuff
        s_i = steps['s']
        t_i = steps['t']
        c = self.medium.c
        Nsamples = T.shape[1]
        #set the initial values
        #s_i[0] = src.s
        #rotate the trajectories
        r = align_rotation_to_vector(src.s)
        s_rot = np.asarray([r.apply(s) for s in s_i]).view(vector)
        s_i = s_rot
        #apply the delta-function - calculate the last step
        l_last,s_last, delta_factor = self.calculate_last_step(R,T, t_i, s_i)
        T1 = T-l_last/c
        # --- mask the unphysical values
        mask = (T*c< R.mag()).flatten()
        mask |= (l_last<0).flatten()
        mask |= ((T1<0)|(T1>T)).flatten()
        
        if self.use_masking:
            if np.all(mask):
                return np.zeros((1,Nsamples))
            l_last, s_last, delta_factor = mask_arrays([l_last, s_last, delta_factor], mask=mask, axis=1)
            s_i, t_i, T1, T = mask_arrays([s_i, t_i, T1, T], mask=mask, axis=1)
            
        #gather the resulting values
        s = np.concatenate([s_i, s_last], axis=0).view(vector)
        t = np.concatenate([t_i*T1, T], axis=0)
        #calculate factors
        scat_factor = self.medium.scatter(s[-2].dot(s[-1]))[np.newaxis,:]
        att_factor = self.medium.attenuation(T, n_scattering=self.Nsteps)
        step_factor = (T1/T)**(self.Nsteps-1)
        self.factor = scat_factor*att_factor*step_factor*delta_factor
        self.factor = self.factor.reshape(s.shape[1])
        self.factor*= self.medium.c
        #optional stuff
        if self.use_masking:
            # --- unmask all arrays --- 
            t,s = unmask_arrays([t,s], mask=mask, axis=1)
            self.factor = unmask_arrays(self.factor, mask=mask, axis=0)
        else:
            self.factor[mask]=0
        if self.save_trajectory:
            #calculate the trajectory (for debugging, can be disabled)
            dt = np.diff(t, axis=0)
            dr = c*np.cumsum(dt*s, axis=0)
            r = np.concatenate([np.zeros((1,*dr.shape[1:])),dr],axis=0)+src.R
            self.trajectory = swap_axes(Point(r,t,s))#store the trajectory
        
        #return ones - they will be multiplied by the factor automatically
        return np.ones(shape=(1,Nsamples))

    def __getitem__(self, key):
        expr = super()
        for token in key.split('.'):
            expr = expr.__getitem__(token)
        return expr
    
    def __setitem__(self, key, value):
        try:
            path, key = key.rsplit('.',1)
            expr = self[path]
        except ValueError:
            expr = super()
        expr.__setitem__(key, value)
        
    def calculate(self, override:dict=None):
        """Run the calculation using the vegas integrator.

        Parameters
        ----------
        override: dict[str, value]
            Change some parameters to given values, before calculating. 
            If None (default), just run the calculation.
        Returns
        -------
        gvar.GVar:
            The integration result, containing mean value and sdev.

        Notes
        ------
            * User can change the vegas integrator parameters via `Process.vegas_kwargs` variable
            * Override just replaces the initial values of given parameters. 
            Running 
            ```
            p.calculate(override={'src.T':1})
            ``` 
            is equivalent to
            ```python
            p['src.T']=1
            p.calculate()
            ```
        """
        if override is not None:
            for key, value in override.items():
                if not key.startswith('_'):
                    self[key]=value
        return vp.integral(self)(**self.vegas_kwargs)

    def calculate_map(self, override:dict, 
                      output='dict', 
                      map_function="ProcessPool"):
        """Perform several calculations, for each of the values defined in 'override' dictionary.
        
        Parameters
        ----------
        override : dict[str, list]
            A dictionary where key defines the parameter to be changed, and value defines the list of values for that parameter. 
        
        output : 'dict'|'reshape'
            Expected output shape. 
            If 'dict' (default) - return the dict with values and result
            If 'reshape' - return an array of values, with dimensions corresponding to given parameters.
        
        map_function : callable or "ProcessPool"
            This can be a default python `map` function, or its equivalent.
            If a string "ProcessPool" is given - use `concurrent.futures.ProcessPoolExecutor` to run calculation concurrently.
            
        Returns
        -------
        list of dicts, or np.array, depending on the `output` parameter
        
        Notes
        -----
        If 'override' has several keys, the resulting calculation will be performed for the Cartesian product of these parameters (i.e. `override={'A':[1,2], 'B':[3,4]}` means calculation for `[{'A':1,'B':3},{'A':2,'B':3},{'A':1,'B':4}{'A':2,'B':4}]`
        """
        if map_function == 'ProcessPool':
            with ProcessPoolExecutor() as executor:
                return self.calculate_map(override=override, output=output, map_function=executor.map)
                
        keys,values = override.keys(), override.values()
        kwargs = [dict(zip(keys,v)) for v in product(*values)]
        #run the actual calculation
        result =  list(map_function(self.calculate, kwargs))
        #output the result
        if output=='reshape':
            return np.asarray(result).reshape([len(v) for v in values])
        elif output=='dict':
            return [dict(**d,result=r) for d,r in zip(kwargs,result)]
        else:
            return result


import numpy as np
import vegas_params as vp
from vegas_params.vector import vector
from scipy.spatial.transform import Rotation

def combine_rotations(dir_base:vector, dir_relative:vector)->vector:
    rot = align_rotation_to_vector(dir_base)
    return rot.apply(dir_relative.reshape(-1,3)).view(vector)

def align_rotation_to_vector(v:vector)->Rotation:
    v = np.asarray(v).view(vector)
    n = v/v.mag()
    theta = np.arccos(n.z)
    phi = np.arctan2(n.y,n.x)
    return Rotation.from_euler('yz',np.array([theta,phi]).T.squeeze())



from dataclasses import dataclass
import numpy as np

class HeneyGreenshtein:
    def __init__(self, g=0.9):
        self.g = g
    
    def pdf(self, x:np.ndarray)->np.ndarray:
        "probability density function"
        g = self.g
        return 0.5*(1 - g**2) / ((1 + g**2 - 2*g*x)**1.5)
    
    def cdf(self, x:np.ndarray)->np.ndarray:
        "Cumulative distirbution function"
        g = self.g
        return 0.5*(1-g**2)/g*(1/np.sqrt(1 + g**2 - 2*g*x) - 1/(1+g))

    def icdf(self, y:np.ndarray):
        "Inverse of cdf"
        g = self.g
        return (1+g**2 - ((1-g**2)/(2*y*g+1-g))**2)/(2*g)

@dataclass
class Medium:
    mu_a:float #inverse absorption length
    mu_s:float #inverse scattering length
    c:float  #speed of light in the medium
    g:float # average cosine of scattering (asymmetry parameter)
    
    def __post_init__(self):
        self.mu_t=self.mu_a+self.mu_s
        self.h_g = HeneyGreenshtein(self.g)

    def attenuation_factor(self, time:np.ndarray)->np.ndarray:
        return np.exp(-self.mu_t*self.c*time)

    def attenuation(self, time:np.ndarray, n_scattering:int)->np.ndarray:
        return ((self.c*self.mu_s*time)**n_scattering) * self.attenuation_factor(time)

    def scatter(self, x:np.ndarray)->np.ndarray:
        return self.h_g.pdf(x)/(2*np.pi)

#define the medium parameters
water = Medium(mu_a = 0.047846, mu_s = 0.014439, c = 219662673.8, g = 0.9)


import numpy as np
from vegas_params import expression, Uniform
from vegas_params.vector import vector, Vector, Scalar, Direction
from dataclasses import dataclass

from .utils import combine_rotations

@dataclass 
class Point:
    """A container for the photon trajectory points"""
    R: vector
    T: np.array
    s: vector
    def __getitem__(self, n):
        return Point(self.R[n],self.T[n],self.s[n])
    def __len__(self):
        return len(np.atleast_1d(self.R))
    def iter(self):
        for n in range(len(self)):
            yield self[n]

@expression
class Spherical:
    r:Scalar = Scalar(Uniform([0,1]))
    s:Direction = Direction()
    def __call__(self, r,s):
        self.factor=r**2
        return r*s
        
@expression
class Target:
    R: Vector = Spherical()
    T: Scalar = Uniform([0,1])
    s: Direction = Direction(1,0)#fixed: not used
    
    def __call__(self, R, T, s):
        self.factor=1
        return Point(np.atleast_3d(R),
                     np.atleast_3d(T),
                     np.atleast_3d(s))

class Source(Target):
    pass

@expression
class TrackSource:
    T: Scalar
    track_pos: Vector = Vector([0,0,0])
    track_speed: Scalar = 3e8 #m/s
    track_dir: Direction = Direction(1,0)
    photon_dir: Direction = Direction(cos_theta=0.9)

    @staticmethod
    def __call__(T, track_pos, track_speed, track_dir, photon_dir):
        R = track_pos+track_dir*track_speed*T
        s = combine_rotations(track_dir, photon_dir)
        return Point(R,T,s)


import pyqtgraph as pg
import pyqtgraph.opengl as gl
import numpy as np

def normalize(a, axis=None, v_min=None, v_max=None):
    v_min = v_min or a.min(axis=axis)
    v_max = v_max or a.max(axis=axis)
    if v_min==v_max:
        return np.ones_like(a)
    return (a - v_min)/(v_max-v_min)

def plot_trajectory(points, factor):
    view = gl.GLViewWidget()
    #grids
    #view.addItem(grid:=gl.GLGridItem())
    #axes
    view.addItem(zaxis:=gl.GLGridItem())
    zaxis.setSize(0,0,10)
    #colormap
    points.T = normalize(points.T, axis=None).squeeze()
    cm = pg.colormap.get('CET-D8')
    for pt, f in zip(points.iter(),normalize(factor)):
        width = 1+f*3
        alpha = f**0.3
        #logger.debug(p.R.shape)
        if width<0.75 or alpha<0.1:
            continue
        color = cm.map(1-pt.T, mode='Float')
        color[:,3]=alpha#set the alpha
        
        line=gl.GLLinePlotItem(pos=pt.R, 
                               color=color,
                               width=width,
                               antialias=False)
        view.addItem(line)
    view.show()

def plot_photons(points, photon_size=1):
    view = gl.GLViewWidget()
    #grids
    #view.addItem(grid:=gl.GLGridItem())
    #axes
    view.addItem(zaxis:=gl.GLGridItem())
    zaxis.setSize(0,0,10)
    #colormap
    alpha=0.7
    width=2
    factor=1
    points.T = normalize(points.T, axis=None).squeeze()
    cm = pg.colormap.get('CET-D8')
    for pt in points.iter():
        color = cm.map([1-pt.T]*2, mode='Float')
        #color[:,3]=alpha#set the alpha
        line=gl.GLLinePlotItem(pos=[pt.R, pt.R+pt.s*photon_size], 
                               color=color,
                               width=width,
                               antialias=False)
        #plot all the photon directions
        view.addItem(line)
    view.show()


import numpy as np
from scipy.spatial.transform import Rotation

import vegas_params as vp
from vegas_params.vector import vector
from vegas_params.utils import swapaxes
from .medium import Medium

from loguru import logger

#intermediate steps

#times
@vp.expression
@swapaxes
def Times_from_xi(self, xi):
    r"""Construct normalized times (t/T) from given xi values:

    Parameters:
    -----------
        xi: np.ndarray
            Array of shape (Nsamples,N,...) of values :math:`\xi\in[0,1]`
    Factor:
    -------
    .. math ::
        f = \prod_{i=1}^{N} (1-\xi_i)^{i-1}

    Returns: 
    --------
    array of normalized time of shape (Nsamples,N,...), t[0]=0, t[N]=1
    .. math ::
        \tau_i = \prod_{k=i}^{N}(1-\xi_k)
    """
    #calculate factor
    powers = np.arange(len(xi))-1
    powers = powers.reshape(-1,1)
    self.factor = np.prod(np.power((1.-xi)[1:],powers[1:]), axis=0)
    #calculate T
    xi_rev = np.flip(xi, axis=0)
    t_rev = np.cumprod(1-xi_rev, axis=0)
    t = np.flip(t_rev, axis=0)
    t = np.atleast_3d(t)
    #append last ending point
    t = np.append(t, np.ones(shape=(1,*t.shape[1:])), axis=0)
    #return result
    return t

#directions
class Directions_from_scattering_distr(vp.Expression):
    def __init__(self, medium, Nsteps:int, s0=vp.Direction(1,0)):
        self.medium = medium
        self.Nsteps = Nsteps
        super().__init__(s0=s0, 
                         #cos_theta = Fixed([0.5]*(Nsteps-1)),
                         cos_theta = vp.FromDistribution(medium.h_g.icdf)**(Nsteps-1),
                         phi = vp.Uniform([0,2*np.pi])**(Nsteps-1)
                        )
    @swapaxes
    def __call__(self, s0, cos_theta, phi):
        theta = np.arccos(cos_theta)
        step_dir = s0.reshape(-1,3)
        s = [step_dir]
        for theta_step,phi_step in zip(theta, phi):
            #transversal vector along X(if possible)
            s_T = np.cross(step_dir, [1,0,0]).view(vector)
            #recalculate for Y axis, if the cross with X axis was zero
            is_zero = np.isclose(s_T.mag(),0).squeeze()
            s_T[is_zero,:] = np.cross(step_dir[is_zero], [0,1,0])
            #apply the rotation
            rot_vec0 = (phi_step[:,np.newaxis]*step_dir)
            rot_axis = Rotation.from_rotvec(rot_vec0).apply(s_T).view(vector)
            #additionally normalize rot_axis
            rot_axis = rot_axis/rot_axis.mag()
            rot_vec1 = (theta_step*rot_axis)
            #
            step_new = Rotation.from_rotvec(rot_vec1).apply(step_dir) #apply rotation to the previous step
            step_dir = step_new
            s.append(step_new)
        s = np.asarray(s).view(vector)
        return s

#steps variants
class StepsUniform(vp.Expression):
    """Expression that describes the intermediate photon steps"""
    def __init__(self, Nsteps:int, medium:Medium, s0:vp.Direction=vp.Direction(1,0)):
        """
        Parameters
        ----------
        Nsteps:int
            number of intermediate photon steps (Number of scatterings)
        medium:Medium
            scattering medium, to define the probability pistribution
        s0:Expression
            Direction of the first step (default is fixed along positive Z)
        Returns
        -------
            Dictionary:
                't': time of each step, shape=(Nsamples,Nsteps,1),
                's': direction at each step, shape=(Nsamples,Nsteps,3)}
        """
        self.Nsteps = Nsteps
        self.medium = medium
        super().__init__(t=Times_from_xi(xi=1|vp.Uniform([0,1])**(Nsteps-1)),
                         s=s0|vp.Direction()**(Nsteps-1)
                        )
        
    @swapaxes
    def __call__(self, t, s):
        s = s.view(vector)
        s_dot_s = s[1:].dot(s[:-1])
        self.factor = np.prod(self.medium.scatter(s_dot_s), axis=0).swapaxes(0,1)
        return {'t':t, 's':s}

class StepsDistribution(vp.Expression):
    """Expression that describes the intermediate photon steps"""
    def __init__(self, Nsteps:int, medium:Medium, s0:vp.Direction=vp.Direction(1,0)):
  
        self.Nsteps = Nsteps
        super().__init__(t=Times_from_xi(xi=1|vp.Uniform([0,1])**(Nsteps-1)),
                         s=Directions_from_scattering_distr(medium, Nsteps, s0=s0)
                        )
        
    def __call__(self, t, s):
        self.factor = 1/(2*np.pi)**(self.Nsteps-1)
        return {'t':t, 's':s}
            



import vegas_params as vp
import rte
import numpy as np
import pytest


v_expected = {1e-6: 3.34332961e-07, 1e-8:1.10669159e+04}

vegas_kwargs = dict(nitn=10, neval=10000)

@pytest.mark.parametrize(
    'R, T, Nsteps, expected',
    [
        ([1,0,1], 1e-6, 1, 3.34332961e-07),
        ([1,0,1], 1e-8, 1, 1.10669159e+04)
    ])
def test_fixed_time(R, T, Nsteps, expected):
    #initial setup for source and target
    src = rte.Source(R=vp.Vector([0,0,0]),
                     T=0,
                     s=vp.Vector([0,0,1]))
    
    tgt = rte.Target(R=vp.Vector(R),
                     T=T,
                     s=vp.Vector([0,0,1]))

    p = rte.Process(src, tgt, medium=rte.medium.water, Nsteps=Nsteps)
    p.vegas_kwargs = vegas_kwargs
    result = p.calculate()
    assert np.isclose(result.mean, expected)
